/*********************************************************************************************************
* Software License Agreement (BSD License)                                                               *
* Author: Sebastien Decugis <sdecugis@nict.go.jp>							 *
*													 *
* Copyright (c) 2008, WIDE Project and NICT								 *
* All rights reserved.											 *
* 													 *
* Redistribution and use of this software in source and binary forms, with or without modification, are  *
* permitted provided that the following conditions are met:						 *
* 													 *
* * Redistributions of source code must retain the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer.										 *
*    													 *
* * Redistributions in binary form must reproduce the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer in the documentation and/or other						 *
*   materials provided with the distribution.								 *
* 													 *
* * Neither the name of the WIDE Project or NICT nor the 						 *
*   names of its contributors may be used to endorse or 						 *
*   promote products derived from this software without 						 *
*   specific prior written permission of WIDE Project and 						 *
*   NICT.												 *
* 													 *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED *
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 	 *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 	 *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR *
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF   *
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.								 *
*********************************************************************************************************/

/* This file contains all support functions to parse, create, and manipulate RADIUS messages. Other 
modules do not need to "know" the actual representation of RADIUS messages on the network. They only 
receive the logical view as exposed in the radius_gw.h file. */

#include "radius_gw_internal.h"

/* To ensure packed structures with some common compilers */
#ifdef __GNUC__
#define STRUCT_PACKED __attribute__ ((packed))
#else
#define STRUCT_PACKED
#endif
#ifdef _MSC_VER
#pragma pack(push, 1)
#endif /* _MSC_VER */

struct radius_hdr {
	uint8_t code;
	uint8_t identifier;
	uint16_t length; /* including this header */
	uint8_t authenticator[16];
	/* followed by length-20 octets of attributes */
} STRUCT_PACKED;

struct radius_attr_hdr {
	uint8_t type;
	uint8_t length; /* including this header */
	/* followed by length-2 octets of attribute value */
} STRUCT_PACKED;

#ifdef _MSC_VER
#pragma pack(pop)
#endif /* _MSC_VER */

int rgw_msg_parse(unsigned char *buf, size_t len, rad_t **msg)
{
	struct radius_hdr * hdr = (struct radius_hdr *)buf;
	rad_t * new = NULL;
	size_t remaining = 0;
	struct radius_attr_hdr *next;
	
	TRACE_ENTRY("%p %g %p", buf, len, msg);
	
	CHECK_PARAMS(buf && msg);
	
	if (len < sizeof(struct radius_hdr) ) {
		TRACE_DEBUG(FULL, "Message too short received (%g bytes), discard.\n", len);
		return EINVAL;
	}
	
	remaining = ntohs(hdr->length);
	if (len < remaining) {
		TRACE_DEBUG(FULL, "Truncated message received (%g / %g bytes), discard.\n", len, remaining);
		return EINVAL;
	}
	
	if (len > remaining) {
		TRACE_DEBUG(FULL, "Ignore extra bytes at the end of RADIUS message (%g / %g bytes).\n", len, remaining);
	}
	
	/* Ok, we can try and parse the buffer */
	CHECK_MALLOC( new = malloc(sizeof(rad_t)) );
	memset(new, 0, sizeof(rad_t));
	rg_list_init(&new->attributes);
	
	/* Copy header data */
	new->code = hdr->code;
	new->identifier = hdr->identifier;
	memcpy(&new->authenticator[0], &hdr->authenticator[0], 16);
	
	/* Parse the attributes */
	next = (struct radius_attr_hdr *)(hdr + 1);
	remaining -= sizeof(struct radius_hdr);
	
	while (remaining > sizeof(struct radius_attr_hdr)) {
		struct rad_attr * newattr = NULL;
		struct radius_attr_hdr *cur = next;
		
		if ((cur->length > remaining) || (cur->length < sizeof(struct radius_attr_hdr)))
			break;

		next = (struct radius_attr_hdr *)(((unsigned char *)cur) + cur->length);
		remaining -= cur->length;
		
		CHECK_MALLOC( newattr = malloc(sizeof(struct rad_attr)) );
		memset(newattr, 0, sizeof(struct rad_attr));
		rg_list_init(&newattr->chain);
		
		newattr->type = cur->type;
		newattr->length = cur->length;
		
		if ( cur->length > sizeof(struct radius_attr_hdr) )
			memcpy(&newattr->data.buf[0], cur+1, cur->length - sizeof(struct radius_attr_hdr));
		
		rg_list_insert_before(&new->attributes, &newattr->chain);
	}
	
	/* Done! */
	*msg = new;
	return 0;
}

int rgw_msg_gen(rad_t *msg, unsigned char **buf, size_t *len)
{
	size_t mylen;
	struct rg_list * li;
	unsigned char * mybuf = NULL;
	struct radius_hdr * hdr;
	struct radius_attr_hdr *attr;
	
	TRACE_ENTRY("%p %p %p", msg, buf, len);
	
	CHECK_PARAMS(msg && buf && len);
	
	/* Compute the size of the final message */
	mylen = sizeof(struct radius_hdr);
	
	for (li = msg->attributes.next; li != &msg->attributes; li = li->next)
		mylen += ((struct rad_attr *)li)->length;
	
	CHECK_PARAMS( mylen < (1<<16) );
	
	CHECK_MALLOC( mybuf = malloc(mylen) );
	memset(mybuf, 0, mylen);
	
	/* Now write the header */
	hdr = (struct radius_hdr *)mybuf;
	hdr->code = msg->code;
	hdr->identifier = msg->identifier;
	hdr->length = htons( (unsigned short)mylen );
	memcpy(&hdr->authenticator[0], &msg->authenticator[0], 16);
	
	attr = (struct radius_attr_hdr *)(hdr+1);
	for (li = msg->attributes.next; li != &msg->attributes; li = li->next) {
		struct rad_attr * attrmem = (struct rad_attr *)li;
		
		attr->type = attrmem->type;
		attr->length = attrmem->length;
		
		if (attr->length > sizeof(struct radius_attr_hdr)) {
			memcpy(attr+1, &attrmem->data.buf[0], attr->length - sizeof(struct radius_attr_hdr));
		}
		
		attr = (struct radius_attr_hdr *)(((unsigned char *)attr) + attrmem->length);
	}
	
	/* Done */
	*buf = mybuf;
	*len = mylen;
	return 0;
}


