/*********************************************************************************************************
* Software License Agreement (BSD License)                                                               *
* Author: Sebastien Decugis <sdecugis@nict.go.jp>							 *
*													 *
* Copyright (c) 2008, WIDE Project and NICT								 *
* All rights reserved.											 *
* 													 *
* Redistribution and use of this software in source and binary forms, with or without modification, are  *
* permitted provided that the following conditions are met:						 *
* 													 *
* * Redistributions of source code must retain the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer.										 *
*    													 *
* * Redistributions in binary form must reproduce the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer in the documentation and/or other						 *
*   materials provided with the distribution.								 *
* 													 *
* * Neither the name of the WIDE Project or NICT nor the 						 *
*   names of its contributors may be used to endorse or 						 *
*   promote products derived from this software without 						 *
*   specific prior written permission of WIDE Project and 						 *
*   NICT.												 *
* 													 *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED *
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 	 *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 	 *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR *
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF   *
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.								 *
*********************************************************************************************************/

/* This file contains the definitions of types and functions involved in the routing decisions in waaad,
 * and that can be called by extensions. 
 *
 * Three different processing can be distinguished:
 *  - Messages received, and the peer is final recipient (IN messages)
 *  - Messages received, and the peer is not final recipient (FWD messages)
 *  - Message is locally generated (OUT messages)
 *
 * There are three global message queues (declared in queues.h) and peers-specific queues (in peer.c).
 *
 * (*) IN messages processing details:
 *   - the message is received in a peer.
 *   - after some checks, the message is put in the global "incoming" queue, with associated metadata.
 *   - a daemon thread (routing-in) makes the decision if the message is handled locally or forwarded, 
 *     based on local capabilities (registered by extensions).
 *   - If the message is handled locally, it is put in the global "local" queue.
 *   - Another thread (dispatch) will handle this message and pass it to registered callbacks (see handle-api.h).
 *
 * (*) FWD messages details:
 *   - The process is the same as for IN messages, until the routing-in threads makes its decision.
 *   - If the message is to be forwarded, rt_fwd_cb_t callbacks are called.
 *     - these callbacks will do the proxying work. Note that adding the route-record is handled by the daemon.
 *   - Once processed, the message is put in the global "outgoing" queue.
 *   - The remaining process is the same as for OUT messages, as described bellow.
 *
 * (*) OUT messages details:
 *   - Once created, the message is put in the global "outgoing" queue (or is a forwarded message).
 *   - a thread will pick the message, and build a list of possible destinations for it.
 *     The logic to build this list is as follow:
 *      - create a list of all peers in the "OPEN" state.
 *      - remove all peers that are already in a Route-Record of the message, to avoid routing loops.
 *   - If the list has zero peer remaining, create an error UNABLE_TO_DELIVER.
 *   - If the list has only one peer, send the message to this peer.
 *   - Otherwise, call all registered rt_out_cb_t callbacks (registered by daemon and extensions) with the list of peers.
 *   - Order the resulting list of peers by score, and sent the message to the peer with highest score.
 *    - in case the peer is no longer in the "OPEN" state, send the message to the second peer in the list.
 *      - if no peer is in OPEN state anymore, restart the process of creating the list.
 *   - The peer thread will handle the creation of the Hop-by-hop ID and sending the message.
 *
 * This part of the API (routing-api.h) provides the definitions of the rt_out_cb_t and rt_fwd_cb_t callbacks, and the
 * functions to register and deregister these callbacks.
 */
 


#ifndef _ROUTING_API_H
#define _ROUTING_API_H

#include <waaad/message-api.h>

/******************************************************************
 *   The following part is dedicated to the OUT messages handling *
 * (see introduction at the top of this file)                     *
 ******************************************************************/

/* The following values are typical "scores" that extensions should use in routing decisions */
#define	PEER_SCORE_FINALDEST		 100	/* If the peer is the final recipient of the message, it receives a big score. */
#define PEER_SCORE_REDIR_SESSION	  50	/* If there is a redirect rule with ALL_SESSION for these message and peer */
#define PEER_SCORE_REDIR_USER	 	  45	/* If there is a redirect rule with ALL_USER for these message and peer */
#define PEER_SCORE_REDIR_REALM_APP 	  40	/* If there is a redirect rule with REALM_AND_APPLICATION for these message and peer */
#define PEER_SCORE_REDIR_REALM 		  35	/* If there is a redirect rule with ALL_REALM for these message and peer */
#define PEER_SCORE_REDIR_APP 		  30	/* If there is a redirect rule with ALL_APPLICATION for these message and peer */
#define PEER_SCORE_REDIR_HOST 		  25	/* If there is a redirect rule with ALL_HOST for these message and peer */
#define PEER_SCORE_DEFAULT_REALM	  10	/* The peer is a default route for this realm */
#define PEER_SCORE_DEFAULT		   5	/* The peer is a default route for all messages */
#define PEER_SCORE_LOAD_BALANCE		   1	/* Use this to differentiate between several peers with the same score */
#define PEER_SCORE_ERR_BUSY		 -60	/* The peer returned a BUSY error */
#define PEER_SCORE_ERR_UNABLE_DELIV	-200	/* The peer returned a UNABLE_TO_DELIVER error */

/* The following type is used to represent a dynamic list of peers to which the message could be sent. */
typedef struct _rt_dpl {
	peer_t		*peer;	/* The peer object of this entry */
	int 	 	 score;	/* The current score of this peer. The functions should not care of the value inside and just add their own score. message will be sent to peers with >0 score only */
	struct _rt_dpl  *next;	/* Pointer to the next candidate peer, or NULL for the last element */
} rt_dpl_t;


/* The following is the prototype of a function that handles routing decisions for outgoing messages */
/*
 * CALLBACK:	rt_out_cb_t
 *
 * PARAMETERS:
 *  data	: pointer to some data that was passed when the callback was registered.
 *  msg 	: The message that must be sent.
 *  list        : The list of peers to which the message may be sent (peers in active state and no routing loop)
 *
 * DESCRIPTION: 
 *   This callback will attribute a score (preferably from PEER_SCORE_*) to each peer in the list.
 *  When all registered callbacks have terminated, the message is sent to the peer with the highest score.
 *  Note that each callback must *add* its locally-attributed score to the "score" parameter, not replace it!
 *  Note also that this callback must be re-entrant since it may be called by several threads at the same time 
 *  (on different messages)
 *
 * RETURN VALUE:
 *  0      	: Operation complete.
 *  !0 		: An error occurred.
 */
typedef int (*rt_out_cb_t) ( void * data, msg_t * msg, rt_dpl_t * list );

/* The following enum represents the position in which this callback should be registered */
typedef enum {
	RT_OUT_NORMAL,	/* This callback can be called at any time */
	RT_OUT_LATE	/* This callback should be called as late as possible in the callback list */
} rt_out_pos_t;

/* The following opaque type represents a handler to a registered callback */
typedef void rt_out_hdl_t;

#ifndef IN_EXTENSION

/*
 * FUNCTION:	rt_out_register
 *
 * PARAMETERS:
 *  cb 		  : The callback function to register.
 *  data	  : Pointer to pass to the callback when it is called.
 *  pos           : When this callback must be called.
 *  handler       : On success, a handler to the registered callback is stored here. 
 *		   This handler will be used to unregister the cb.
 *
 * DESCRIPTION: 
 *   Register a new callback to handle OUT routing decisions.
 *
 * RETURN VALUE:
 *  0      	: The callback is registered.
 *  EINVAL 	: A parameter is invalid.
 *  ENOMEM	: Not enough memory to complete the operation
 */
int rt_out_register ( rt_out_cb_t cb, void * data, rt_out_pos_t pos, rt_out_hdl_t ** handler );

/*
 * FUNCTION:	rt_out_unregister
 *
 * PARAMETERS:
 *  handler       : The callback that must be unregistered.
 *
 * DESCRIPTION: 
 *   Removes a callback from the list of registered callbacks.
 *
 * RETURN VALUE:
 *  0      	: The callback is unregistered.
 *  EINVAL 	: A parameter is invalid.
 */
int rt_out_unregister ( rt_out_hdl_t * handler );

#endif /* ! IN_EXTENSION */


/******************************************************************
 *   The following part is dedicated to the FWD messages handling *
 * (see introduction at the top of this file)                     *
 ******************************************************************/

/* The following is the prototype of a function that handles messages transformations of forwarded messages */
/*
 * CALLBACK:	rt_fwd_cb_t
 *
 * PARAMETERS:
 *  data	: pointer to some data that was passed when the callback was registered.
 *  msg 	: The message that is being forwarded.
 *
 * DESCRIPTION: 
 *   This callback is called when a message is forwarded to another peer. It may for example add a Proxy-Info AVP.
 *
 * RETURN VALUE:
 *  0      	: Operation complete.
 *  !0 		: An error occurred / the message must not be forwarded (example: redirect indications)
 */
typedef int (*rt_fwd_cb_t) ( void * data, msg_t * msg );

/* The following values define what kind of forwarded message the callback must be called with */
typedef enum {
	RT_FWD_REQ = 1,	/* This callback will be called on forwarded requests only */
	RT_FWD_ANS,	/* This callback will be called on answers and errors only */
	RT_FWD_ALL,	/* This callback will be called on all forwarded messages */
} rt_fwd_pos_t;

/* The following opaque type represents a handler to a registered callback */
typedef void rt_fwd_hdl_t;

#ifndef IN_EXTENSION

/*
 * FUNCTION:	rt_fwd_register
 *
 * PARAMETERS:
 *  cb 		  : The callback function to register.
 *  data	  : Pointer to pass to the callback when it is called.
 *  pos           : On which messages this callback must be called.
 *  handler       : On success, a handler to the registered callback is stored here. 
 *		   This handler will be used to unregister the cb.
 *
 * DESCRIPTION: 
 *   Register a new callback to handle FWD messages.
 *
 * RETURN VALUE:
 *  0      	: The callback is registered.
 *  EINVAL 	: A parameter is invalid.
 *  ENOMEM	: Not enough memory to complete the operation
 */
int rt_fwd_register ( rt_fwd_cb_t cb, void * data, rt_fwd_pos_t pos, rt_fwd_hdl_t ** handler );


/*
 * FUNCTION:	rt_fwd_unregister
 *
 * PARAMETERS:
 *  handler       : The callback that must be unregistered.
 *
 * DESCRIPTION: 
 *   Removes a callback from the list of registered callbacks.
 *
 * RETURN VALUE:
 *  0      	: The callback is unregistered.
 *  EINVAL 	: A parameter is invalid.
 */
int rt_fwd_unregister ( rt_fwd_hdl_t * handler );

#endif /* ! IN_EXTENSION */





/******************************************************************/


/* The version of this API, to check binary compatibility -- increment each time a change is made in api_rt_t */
#define WAAAD_API_RT_VER	1

/* Now define the type of the structure that contains the callback to pass to extensions */
typedef struct {
	/* The header is common to all sub-API pieces */
	size_t	length;		/* The size of this structure, may be useful for extensions not using the facility */
	int	version;	/* The version of this API/ABI, must be WAAAD_API_RT_VER */
	
	/* the remaining is routing-specific */
	int (*rt_out_register)    ( rt_out_cb_t cb, void * data, rt_out_pos_t pos, rt_out_hdl_t ** handler );
	int (*rt_out_unregister)  ( rt_out_hdl_t * handler );
	int (*rt_fwd_register)    ( rt_fwd_cb_t cb, void * data, rt_fwd_pos_t pos, rt_fwd_hdl_t ** handler );
	int (*rt_fwd_unregister)  ( rt_fwd_hdl_t * handler );
} api_rt_t;

#ifdef IN_EXTENSION

/* From within the extensions, we register callbacks in the following global structure */
#ifdef DECLARE_API_POINTERS
api_rt_t * g_api_rt=NULL;
#else /* DECLARE_API_POINTERS */
extern api_rt_t * g_api_rt;
#endif /* DECLARE_API_POINTERS */

/* These defines allow to call functions from extension in the same way as in waaad */
#define rt_out_register		g_api_rt->rt_out_register
#define rt_out_unregister	g_api_rt->rt_out_unregister
#define rt_fwd_register		g_api_rt->rt_fwd_register
#define rt_fwd_unregister	g_api_rt->rt_fwd_unregister

#else /* IN_EXTENSION */

/* From the daemon, we must initialize the API object, in extension.c */
# define MY_WAAAD_API_RT_VER 1
# if MY_WAAAD_API_RT_VER != WAAAD_API_RT_VER
#  error "You must update API_INIT_RT also"
# endif

#define API_INIT_RT( api_rt ) 				\
{							\
	(api_rt).length             = sizeof(api_rt_t);	\
	(api_rt).version            = WAAAD_API_RT_VER;	\
	(api_rt).rt_out_register    = rt_out_register;	\
	(api_rt).rt_out_unregister  = rt_out_unregister;\
	(api_rt).rt_fwd_register    = rt_fwd_register;	\
	(api_rt).rt_fwd_unregister  = rt_fwd_unregister;\
}							

#endif /* IN_EXTENSION */

#endif /* _ROUTING_API_H */
